/*
** Module   : MenuChuiImpl.java
** Abstract : CHUI client implementation of Menu.
**
** Copyright (c) 2015, Golden Code Development Corporation.
** ALL RIGHTS RESERVED. Use is subject to license terms.
**
**           Golden Code Development Corporation
**                      CONFIDENTIAL
**
** -#- -I- --Date-- -----------------------------Description-----------------------------
** 001 VIG 20150116 Created initial version.
** 002 VIG 20150225 Added event processing and drawing methods.
*/
package com.goldencode.p2j.ui.client.chui;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.List;

import com.goldencode.p2j.ui.Color;
import com.goldencode.p2j.ui.Keyboard;
import com.goldencode.p2j.ui.SubMenuConfig;
import com.goldencode.p2j.ui.WidgetConfig;
import com.goldencode.p2j.ui.WidgetId;
import com.goldencode.p2j.ui.chui.ThinClient;
import com.goldencode.p2j.ui.client.Dimension;
import com.goldencode.p2j.ui.client.Menu;
import com.goldencode.p2j.ui.client.MenuItem;
import com.goldencode.p2j.ui.client.Point;
import com.goldencode.p2j.ui.client.Rectangle;
import com.goldencode.p2j.ui.client.SubMenu;
import com.goldencode.p2j.ui.client.border.Border;
import com.goldencode.p2j.ui.client.border.LineBorder;
import com.goldencode.p2j.ui.client.chui.driver.ChuiOutputManager;
import com.goldencode.p2j.ui.client.event.EventManager;
import com.goldencode.p2j.ui.client.event.Key;
import com.goldencode.p2j.ui.client.event.KeyInput;
import com.goldencode.p2j.ui.client.event.PaintEvent;
import com.goldencode.p2j.ui.client.widget.AbstractContainer;
import com.goldencode.p2j.ui.client.widget.AbstractWidget;
import com.goldencode.p2j.ui.client.widget.Container;
import com.goldencode.p2j.ui.client.widget.Widget;

/**
 * <code>Sub-menu</code> CHUI client implementation class. 
 */
@SuppressWarnings("unchecked")
public class SubMenuChuiImpl
extends SubMenu<ChuiOutputManager>
{
   /** Border around the panel. */
   private Border<ChuiOutputManager> border = null;
   
   /**
    * Special constructor used to restore component using given configuration.
    * 
    * @param   id
    *          The widget's ID.
    * @param   cfg
    *          Configuration data.
    */
   public SubMenuChuiImpl(WidgetId id, WidgetConfig cfg)
   {
      super(id, (SubMenuConfig) cfg);
      
      border = new LineBorder()
      {
         /**
          * Draw line border.
          * 
          * @param   widget
          *          Base widget. 
          * @param   size
          *          Border dimensions.
          */
         @Override
         public void drawBorder(Widget<ChuiOutputManager> widget, Dimension size)
         {
            widget.screen().box(bodyLocation(), 
                     size, 
                     widget.screen().borderColor(widget));
         }
      };
      
      //body = new BorderedPanel(new BorderLayout(), border);
   }
   
   /**
    * Drawing worker method.
    */
   public void drawPanel()
   {
      screen().clear(this);

      if (border != null)
         border.drawBorder(this, dimension());
   }

   /**
    * Lay out container.
    */
   @Override
   public void doLayout()
   {
      super.doLayout();

      Widget<ChuiOutputManager>[] w = widgets();

      for (int i = 0; i < w.length; i++)
      {
         w[i].setLocation(new Point(1, i));

         if (w[i] instanceof SubMenu)
         {
            ((SubMenu<?>) w[i]).doLayout();
         }
      }
   }
   
   /**
    * Draw the peer container. The methods calls the draw() method for
    * ScrollPane instance for the peer container. In addition the required
    * clipping rectangle is calculated for ChildFrame containers because they
    * must be rendered inside the parent container.
    * <p>
    * Drawing for the redirected terminal differs. Labels/headers are drawn
    * only once per page, instead of the usual once per menu.
    */
   @Override
   public void draw()
   {
      Rectangle rect = bounds().translate(parent().screenLocation());
      
      drawTitle(new Point(rect.left(), rect.top()));

      if (config().displayed)
      {
         drawPanel();
         super.draw();
         
         /*
         for (Widget<?> c : widgets())
         {
            if (!c.isVisible())
               continue;
            
            if (c instanceof SubMenu && ((SubMenu<?>)c).config().displayed)
            {
               // c.repaint(new Rectangle(new Point(20,1), new Dimension(20, 5)));
               c.repaint();
            }
         }
         */
      }
   }
   
   /**
    * Hide/show widget.
    * <p>
    * If the menu is being shown and not yet attached to a window, attached it to the default
    * window.
    * 
    * @param   visible
    *          New visible state.
    */
   @Override
   public void setVisible(boolean visible)
   {
      // SUB-MENU is not displayed (dropped down) by default
      // setDisplayed(visible);
      
      for(Widget<ChuiOutputManager> w : widgets())
      {
         w.setVisible(true);
      }

      super.setVisible(visible);
   }

   /**
    * Replace original version to take into account P2J specific behavior.
    * 
    * @see  com.goldencode.p2j.ui.client.widget.Widget#dimension()
    */
   @Override
   public Dimension dimension()
   {
      // TODO take note about toggle boxes and submenus. Probably in menu-item or in sub-menu.
      if (!config().displayed)
      {
         return new Dimension(width(), 1); 
      }
      else
      {
         double width = 0; 
         for(Widget<?> w : widgets())
         {
            AbstractWidget<?> aw = (AbstractWidget<?>) w;
            if (width < aw.width())
            {
               width = aw.width(); 
            }
         }

         return new Dimension(width + 4, widgets().length + 2); // borders and title
      }
   }

   /**
    * Get widget location relative to container.
    * 
    * @return  relative widget location.
    */
   /*@Override
   public Point location()
   {
      return calcLocation();
   }*/

   /**
    * Get absolute location of widget in screen coordinates.
    * 
    * @return  absolute widget location.
    */
   @Override
   public Point screenLocation()
   {
      Point bodyLoc = bodyLocation();
      return new Point(bodyLoc.x + 1, bodyLoc.y + 1);
   }
   
   /**
    * Get widget width.
    * 
    * @return  widget width.
    */
   @Override
   public double width()
   {
      boolean menubar = isMenubar();
      if (!config().displayed && !menubar)
      {
         return config().title.length() + 3; // " ->" will be added
      }
      else if (menubar)
      {
         return config().title.length();
      } 
      else // displayed==true  
      {
         return super.width();
      }
   }

   /**
    * Get widget height.
    * 
    * @return  widget height.
    */
   @Override
   public double height()
   {
      if (!config().displayed)
         return 1;
      else
      {
         // TODO process this case like MENU ?
         // return super.height();
         return getMaxHeight(this);
      }
   }

   /**
    * Calculates a max height of specified Menu or SubMenu.
    * Analyzes nested dropped-down sub-menus.
    * 
    * @param   menu
    *          Menu or SubMenu.
    * 
    * @return  a height of menu.
    */
   static double getMaxHeight(AbstractContainer<?> menu)
   {
      Widget<?>[] widgets = menu.widgets();
      double height = widgets.length + 2;
      for (int i = 0; i < widgets.length; i++)
      {
         if (widgets[i] instanceof SubMenu)
         {
            SubMenu<?> sm = (SubMenu<?>) widgets[i];
            if (sm.config().displayed)
            {
               height = Math.max(height, sm.height() + i + 1);
            }
         }
      }
      return height;
   }
   
   /**
    * Check if widget needs in cursor to be on when having the input focus.  By default this
    * method always returns <code>true</code> so it needs to be overwritten if other return value
    * is required for particular widget implementation class.
    * 
    * @return  <code>true</code> as the default value.
    */
   @Override
   public boolean forceCursorOnFocus()
   {
      return false;
   }

   /**
    * Check if the current widget can manage non-global system actions on its own, and not via the
    * enclosing event loop.
    * 
    * @return   <code>true</code> as the widget has its own event management.
    */
   @Override
   public boolean manageSystemActions()
   {
      return true;
   }
   
   /**
    * Request focus.
    */
   @Override
   public void requestFocus()
   {
      super.requestFocus();
   }
   
   /**
    * Process key event.
    * 
    * @param   event
    *          Key event to process.
    */
   @Override
   public void processKeyEvent(KeyInput event)
   {
      if (event.isConsumed())
      {
         return;
      }

      int key = event.keyCode();
      int actionCode = event.actionCode();
      
      if (key == Key.VK_ENTER || 
          (isMenubar() && actionCode == Keyboard.KA_CURSOR_DOWN) ||
          (!isMenubar() && actionCode == Keyboard.KA_CURSOR_RIGHT))
      {
         // TODO what if body is already displayed
         
         showBody();
         return;
      }
      
      if ((isMenubar() && actionCode == Keyboard.KA_CURSOR_RIGHT) || 
          (!isMenubar() && actionCode == Keyboard.KA_CURSOR_DOWN))
      {
         ThinClient.nextFocus(this);
         if (config().displayed)
         {
            hideBody();
         }
         return;
      }
      
      if ((isMenubar() && actionCode == Keyboard.KA_CURSOR_LEFT) || 
          (!isMenubar() && actionCode == Keyboard.KA_CURSOR_UP))
      {
         ThinClient.previousFocus(this);
         if (config().displayed)
         {
            hideBody();
         }
         return;
      }
      
      // mnemonic
      if (event.keyChar() != KeyInput.CHAR_UNDEFINED)
      {
         parent().processKeyEvent(event);
      }
      else
      {
         super.processKeyEvent(event);
      }
      
   }

   /**
    * Prepare widgets before drawing
    * 
    * @param   widgets
    *          Container children.
    *          
    * @return  widgets, prepared for drawing.
    */
   @Override
   protected List<Widget<ChuiOutputManager>> prepareWidgets(List<Widget<ChuiOutputManager>> widgets)
   {
      List<Widget<ChuiOutputManager>> result = new ArrayList<>();
      result.addAll(widgets);
      Collections.sort(widgets, new Comparator<Widget<ChuiOutputManager>>()
      {
         @Override
         public int compare(Widget<ChuiOutputManager> o1, Widget<ChuiOutputManager> o2)
         {
            if (o1 instanceof MenuItem) return -1;
            if (o2 instanceof SubMenu) return 1;
            return 0;
         }
      });
      return Collections.unmodifiableList(result);
   }
   
   /**
    * Get location of SUB-MENU body, relative to container.
    * 
    * @return  body location.
    */
   private Point bodyLocation()
   {
      Point loc = parent().screenLocation().translate(location());
      return new Point(loc.x, loc.y + 1); 
   }

   /**
    * Repaints SUB-MENU body.
    */
   private void repaintBody()
   {
      Rectangle bodyRect = new Rectangle(bodyLocation(), dimension());
      EventManager.postEvent(new PaintEvent(this, bodyRect));
      
      // repaint();
   }
   
   /**
    * Draws menu-item title, drawing mnemonic character as underlined.
    * 
    * @param   p
    *          Title point.  
    * @param   dColor
    *          Color of the title.
    */
   private void drawTitle(Point p)
   {
      Color dColor = hasFocus() ? Color.MESSAGES : Color.NORMAL; // highlighted() ?
      Color mnemonicColor = Color.INPUT;
      
      String title = this.config().title;
      int ampIdx = title.indexOf('&');
      boolean amp = ampIdx != -1;
      // if & is not specified use the first letter as mnemonic
      if (!amp) ampIdx = 0;
      
      Point tp = new Point(p);
      
      String leftPart = title.substring(0, ampIdx);
      screen().at(tp).append(leftPart, dColor);

      tp.x += ampIdx;
      // "eat" the ampersand
      if (amp) ampIdx++;
      String mnemonic = title.substring(ampIdx, ampIdx + 1);
      screen().at(tp).append(mnemonic, mnemonicColor);

      tp.x += 1;
      String rightPart = title.substring(ampIdx + 1);
      screen().at(tp).append(rightPart, dColor);

      tp.x += rightPart.length();
      
      if (!isMenubar())
      {
         screen().at(tp).append(" ->", dColor);
      }
   }

   /**
    * Check if this sub-menu is right inside MENUBAR (first level)
    * 
    * @return  <code>true</code> if yes.
    */
   private boolean isMenubar()
   {
      Container<ChuiOutputManager> parent = parent();
      return parent instanceof Menu && !((Menu<?>)parent).config().popupOnly;
   }

   /**
    * Hides sub-menu body.
    */
   private void hideBody()
   {
      config().displayed = false;
      repaintBody();
   }
   
   /**
    * Shows this sub-menu body.
    */
   private void showBody() 
   {
      config().displayed = true;
      repaintBody();
   }
}
